/**
 * Common interface to the 802.11 Station Control Block (scb)
 * structure.
 * This file aims to encapsulating the scb structure.
 * Copyright 2005-2006, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
*/

#include <wlc_cfg.h>
#include <typedefs.h>
#include <bcmdefs.h>
#include <osl.h>
#include <sbutils.h>
#include <bcmendian.h>
#include <bcmutils.h>
#include <proto/802.11.h>
#include <proto/wpa.h>
#include <sbconfig.h>
#include <pcicfg.h>
#include <bcmsrom.h>
#include <wlioctl.h>
#include <epivers.h>
#include <proto/eapol.h>
#include <bcmwpa.h>

#include <sbhndpio.h>
#include <sbhnddma.h>
#include <hnddma.h>
#include <d11.h>
#include <wlc_rate.h>
#include <wlc_rate_sel.h>
#include <wlc_key.h>
#include <wlc_channel.h>
#include <wlc_pub.h>
#include <wlc_bsscfg.h>
#include <wlc_pio.h>
#include <wlc.h>
#include <wlc_scb.h>
#include <wlc_phy.h>
#include <wl_export.h>
#include <wlc_channel.h>

#define SCB_MAX_CUBBY		10	/* max number of cubbies in container */

/* structure for storing per-cubby client info */
typedef struct cubby_info {
	scb_cubby_init_t	fn_init;	/* fn called during scb malloc */
	scb_cubby_deinit_t	fn_deinit;	/* fn called during scb free */
	scb_cubby_dump_t	fn_dump;	/* fn called during scb dump */
	void			*context;	/* context to be passed to all cb fns */
} cubby_info_t;

/* structure for storing public and private global scb module state */
typedef struct scb_module_priv {
	scb_module_t	scbpub;				/* public part of scb */
	uint16		nscb;				/* total number of allocated scbs */
	struct scb 	*scbhash[MAXBANDS][NSCBHASH];	/* station control block hash */
	uint		scbtotsize;			/* total scb size including container */
	uint 		ncubby;				/* current num of cubbies */
	cubby_info_t	cubby_info[SCB_MAX_CUBBY];	/* cubby client info */
	wlc_pub_t	*pub;				/* public part of wlc */
} scb_module_priv_t;

#define SCBSTATE_PRIV(s)	((scb_module_priv_t *)(s))
#define SCBSTATE_PUB(s)		(&(s)->scbpub)

/* station control block - one per remote MAC address */
struct scb_info {
	struct scb 	scbpub;		/* public portion of scb */
	struct scb_info *hashnext;	/* pointer to next scb under same hash entry */
	struct scb_info	*next;		/* pointer to next allocated scb */
	struct wlcband	*band;		/* pointer to our associated band */
};

/* Helper macro for txpath in scb */
/* A feature in Tx path goes through following states:
 * Unregisterd -> Registered [Global state]
 * Registerd -> Configured -> Active -> Configured [Per-scb state]
 */
/* Is the feature configured? */
#define SCB_TXMOD_CONFIGURED(scb, fid) (scb->tx_path[(fid)].configured)

/* Set the next feature of given feature */
#define SCB_TXMOD_SET(scb, fid, _next_fid) { \
	scb->tx_path[fid].next_tx_fn = wlc->txmod_fns[_next_fid].tx_fn; \
	scb->tx_path[fid].next_handle = wlc->txmod_fns[_next_fid].ctx; \
	}

static struct scb *wlc_scbvictim(wlc_info_t *wlc);

scb_module_t *
BCMINITFN(wlc_scb_attach)(wlc_pub_t *pub)
{
	scb_module_priv_t *scbstate;

	if ((scbstate = MALLOC(pub->osh, sizeof(scb_module_priv_t))) == NULL)
		return NULL;

	bzero((char *)scbstate, sizeof(scb_module_priv_t));
	scbstate->pub = pub;
	scbstate->scbtotsize = sizeof(struct scb_info);


	return SCBSTATE_PUB(scbstate);
}

void
BCMUNINITFN(wlc_scb_detach)(scb_module_t *scbstate)
{
	if (!scbstate)
		return;

	ASSERT(SCBSTATE_PRIV(scbstate)->nscb == 0);

	MFREE(SCBSTATE_PRIV(scbstate)->pub->osh, scbstate, sizeof(scb_module_priv_t));
}

/* Methods for iterating along a list of scb */

/* Direct access to the next */
struct scb *
wlc_scb_getnext(struct scb *scb)
{
	if (scb)
		return (struct scb *)(((struct scb_info *)scb)->next);
	return NULL;
}

/* Initialize an iterator keeping memory of the next scb as it moves along the list */
void
wlc_scb_iterinit(struct scb_iter *scbiter, struct scb *scb)
{
	ASSERT(scbiter);
	/* Prefetch next scb, so caller can free an scb before going on to the next */
	if (scb)
		scbiter->next = (struct scb *)(((struct scb_info *)scb)->next);
	else
		scbiter->next = NULL;
}

/* move the iterator */
struct scb *
wlc_scb_iternext(struct scb_iter *scbiter)
{
	struct scb *scb;
	ASSERT(scbiter);
	if ((scb = scbiter->next))
		scbiter->next = (struct scb *)(((struct scb_info *)scb)->next);
	else
		scbiter->next = NULL;
	return  scb;
}


/*  
 * Accessors, nagative values are errors.
 */

int
wlc_scb_cubby_reserve(wlc_info_t *wlc, uint size, scb_cubby_init_t fn_init,
	scb_cubby_deinit_t fn_deinit, scb_cubby_dump_t fn_dump, void *context)
{
	uint offset;
	scb_module_priv_t *scbstate = SCBSTATE_PRIV(wlc->scbstate);
	cubby_info_t *cubby_info;

	ASSERT(scbstate->nscb == 0);
	ASSERT((scbstate->scbtotsize % PTRSZ) == 0);
	ASSERT(scbstate->ncubby < SCB_MAX_CUBBY);

	if (scbstate->ncubby >= SCB_MAX_CUBBY) {
		ASSERT(0);
		return -1;
	}

	cubby_info = &scbstate->cubby_info[scbstate->ncubby++];
	cubby_info->fn_init = fn_init;
	cubby_info->fn_deinit = fn_deinit;
	cubby_info->fn_dump = fn_dump;
	cubby_info->context = context;

	offset = scbstate->scbtotsize;

	/* roundup to pointer boundary */
	scbstate->scbtotsize = ROUNDUP(scbstate->scbtotsize + size, PTRSZ);

	return offset;
}

struct wlcband *
wlc_scbband(struct scb *scb)
{
	return ((struct scb_info *)scb)->band;
}


struct scb *
wlc_scballoc(wlc_info_t *wlc)
{
	struct scb_info *scbinfo = NULL;
	struct scb *scb = NULL;
	struct scb *oldscb;
	struct rateset *rs;
	scb_module_priv_t *scbstate = SCBSTATE_PRIV(wlc->scbstate);
	cubby_info_t *cubby_info;
	uint i;

	if (scbstate->nscb < MAXSCB && (scbinfo = MALLOC(wlc->pub.osh, scbstate->scbtotsize)))
		scbstate->nscb++;

	if (!scbinfo) {
		/* free the oldest entry */
		if (!(oldscb = wlc_scbvictim(wlc)))
			return NULL;
		if (!wlc_scbfree(wlc, oldscb))
			return NULL;
		ASSERT(scbstate->nscb < MAXSCB);
		/* allocate memory for scb */
		if (!(scbinfo = MALLOC(wlc->pub.osh, scbstate->scbtotsize)))
			return NULL;
		scbstate->nscb++;
	}

	bzero((char*)scbinfo, scbstate->scbtotsize);
	scb = (struct scb *)scbinfo;

	scb->used = wlc->now;
	scbinfo->band = wlc->band;
	scb->bsscfg = &wlc->cfg;


	for (i = 0; i < scbstate->ncubby; i++) {
		cubby_info = &scbstate->cubby_info[i];
		if (cubby_info->fn_init)
			cubby_info->fn_init(cubby_info->context, scb);
	}

	/* use current, target, or per-band default rateset? */
	if (wlc->pub.up && VALID_CHANNEL(wlc, wlc->target_bss.channel))
		if (wlc->pub.associated)
			rs = &wlc->pub.current_bss.rateset;
		else
			rs = &wlc->target_bss.rateset;
	else
		rs = &wlc->band->defrateset;

	/*
	 * Initialize the per-scb rateset:
	 * - if we are AP, start with only the basic subset of the
	 *	network rates.  It will be updated when receive the next
	 *	probe request or association request.
	 * - if we are IBSS and gmode, special case:
	 *	start with B-only subset of network rates and probe for ofdm rates
	 * - else start with the network rates.
	 *	It will be updated on join attempts.
	 */
	/* initialize the scb rateset */
	if (!wlc->pub.BSS && wlc->band->gmode) {
		wlc_rateset_filter(rs, &scb->rateset, FALSE, TRUE, RATE_MASK);
		/* if resulting set is empty, then take all network rates instead */
		if (scb->rateset.count == 0)
			wlc_rateset_filter(rs, &scb->rateset, FALSE, FALSE, RATE_MASK);
	} else
		wlc_rateset_filter(rs, &scb->rateset, FALSE, FALSE, RATE_MASK);

	wlc_scb_rate_init(wlc, scb);

	/* update scb link list */
	scbinfo->next = (struct scb_info *)SCBSTATE_PUB(scbstate)->scb;
	SCBSTATE_PUB(scbstate)->scb = scb;

	return scb;
}


bool
wlc_scbfree(wlc_info_t *wlc, struct scb *scbd)
{
	int index;
	struct scb_info *remove = (struct scb_info *)scbd;
	struct scb_info *scbinfo;
	scb_module_priv_t *scbstate = SCBSTATE_PRIV(wlc->scbstate);
	cubby_info_t *cubby_info;
	uint i;
	uint8 prio;


	if (scbd->permanent)
		return FALSE;

	for (i = 0; i < scbstate->ncubby; i++) {
		cubby_info = &scbstate->cubby_info[i];
		if (cubby_info->fn_deinit)
			cubby_info->fn_deinit(cubby_info->context, scbd);
	}



	/* free the per station key if one exists */
	if (scbd->key) {
		WL_WSEC(("wl%d: %s: deleting pairwise key for %s\n", wlc->pub.unit,
		        __FUNCTION__, bcm_ether_ntoa(&scbd->ea, eabuf)));
		ASSERT(!bcmp((char*)&scbd->key->ea, (char*)&scbd->ea, ETHER_ADDR_LEN));
		wlc_scb_key_delete(wlc, scbd);

	}

	/* free any frame reassembly buffer */
	for (prio = 0; prio < NUMPRIO; prio++) {
		if (scbd->fragbuf[prio]) {
			PKTFREE(wlc->pub.osh, scbd->fragbuf[prio], FALSE);
			scbd->fragbuf[prio] = NULL;
			scbd->fragresid[prio] = 0;
		}
	}

	scbd->state = 0;

	index = SCBHASHINDEX(scbd->ea.octet);

	/* delete it from the hash */
	scbinfo = (struct scb_info *)(scbstate->scbhash[remove->band->bandunit][index]);
	/* special case for the first */
	if (scbinfo == remove) {
		scbstate->scbhash[remove->band->bandunit][index] = (struct scb *)scbinfo->hashnext;
	}
	else {
		for (; scbinfo; scbinfo = scbinfo->hashnext) {
			if (scbinfo->hashnext == remove)
			{
				scbinfo->hashnext = remove->hashnext;
				break;
			}
		}
		ASSERT(scbinfo);
	}

	/* delete it from the link list */
	scbinfo = (struct scb_info *)(SCBSTATE_PUB(scbstate)->scb);
	if (scbinfo == remove) {
		SCBSTATE_PUB(scbstate)->scb = wlc_scb_getnext(scbd);
	} else {
		while (scbinfo) {
			if (scbinfo->next == remove) {
				scbinfo->next = remove->next;
				break;
			}
			scbinfo = scbinfo->next;
		}
		ASSERT(scbinfo);
	}


	/* free scb memory */
	MFREE(wlc->pub.osh, remove, scbstate->scbtotsize);
	/* update total allocated scb number */
	scbstate->nscb--;

	return TRUE;
}


/* free all scbs, unless permanent. Force indicates reclaim permanent as well */

void
wlc_scbclear(struct wlc_info *wlc, bool force)
{
	struct scb_iter scbiter;
	struct scb *scb;

	if (wlc->scbstate == NULL)
		return;

	FOREACHSCB(&scbiter, wlc->scbstate->scb, scb) {
		if (force)
			scb->permanent = FALSE;
		wlc_scbfree(wlc, scb);
	}
	if (force)
		ASSERT(SCBSTATE_PRIV(wlc->scbstate)->nscb == 0);
}


static struct scb *
wlc_scbvictim(wlc_info_t *wlc)
{
	uint oldest;
	struct scb *scb;
	struct scb *oldscb;
	uint now, age;
	struct scb_iter scbiter;

	if (!wlc->scbstate->scb)
		return NULL;


	/* free the oldest scb */
	now = wlc->now;
	oldest = 0;
	oldscb = NULL;
	FOREACHSCB(&scbiter, wlc->scbstate->scb, scb)  {
		if (!scb->permanent && ((age = (now - scb->used)) >= oldest)) {
			oldest = age;
			oldscb = scb;
		}
	}
	/* handle extreme case: all are permanent */
	if (oldscb == NULL)
		return NULL;

	WL_ASSOC(("scb free %s due to aging %d\n", bcm_ether_ntoa(&oldscb->ea, eabuf),
		oldest));

	return oldscb;
}


void
wlc_scb_setkey(struct scb *scb, wsec_key_t *key)
{
	scb->key = key;
}

/* "|" operation. */

void
wlc_scb_setstatebit(struct scb *scb, uint8 state)
{

	WL_ASSOC(("set state %x\n", state));
	ASSERT(scb);

	if (state & AUTHENTICATED)
	{
		scb->state &= ~PENDING_AUTH;
	}
	if (state & ASSOCIATED)
	{
		ASSERT(scb->state & AUTHENTICATED);
		scb->state &= ~PENDING_ASSOC;
	}
	if (state & AUTHORIZED)
	{
		ASSERT((scb->state & ASSOCIATED));
	}
	scb->state |= state;
	WL_ASSOC(("wlc_scb : state = %x\n", scb->state));
}

/* "& ~" operation */
void
wlc_scb_clearstatebit(struct scb *scb, uint8 state)
{
	ASSERT(scb);
	WL_ASSOC(("clear state %x\n", state));
	scb->state &= ~state;
	WL_ASSOC(("wlc_scb : state = %x\n", scb->state));
}


/* "|" operation . idx = position of the bsscfg in the wlc array of multi ssids.
*/


void
wlc_scb_setstatebit_bsscfg(struct scb *scb, uint8 state, int idx)
{
	ASSERT(scb);
	WL_ASSOC(("set state : %x   bsscfg idx : %d\n", state, idx));
	if (state & ASSOCIATED)
	{

		ASSERT(SCB_AUTHENTICATED_BSSCFG(scb, idx));
		/* clear all bits (idx is set below) */
		memset(&scb->auth_bsscfg, 0, SCB_BSSCFG_BITSIZE);
		scb->state &= ~PENDING_ASSOC;
	}

	if (state & AUTHORIZED)
	{
		ASSERT(SCB_ASSOCIATED_BSSCFG(scb, idx));
	}
	setbit(&scb->auth_bsscfg, idx);
	scb->state |= state;
	WL_ASSOC(("wlc_scb : state = %x\n", scb->state));
}


/*
 * "& ~" operation .
 * idx = position of the bsscfg in the wlc array of multi ssids.
 */

void
wlc_scb_clearstatebit_bsscfg(struct scb *scb, uint8 state, int idx)

{
	int i;
	ASSERT(scb);
	WL_ASSOC(("clear state : %x   bsscfg idx : %d\n", state, idx));
	/*
	   any clear of a stable state should lead to clear a bit
	   Warning though : this implies that, if we want to switch from
	   associated to authenticated, the clear happens before the set
	   otherwise this bit will be clear in authenticated state.
	*/
	if ((state & AUTHENTICATED) || (state & ASSOCIATED) || (state & AUTHORIZED))
	{
		clrbit(&scb->auth_bsscfg, idx);
	}
	/* quik hack .. clear first ... */
	scb->state &= ~state;
	for (i = 0; i < SCB_BSSCFG_BITSIZE; i++)
	{
		/* reset if needed */
		if (scb->auth_bsscfg[i])
		{
			scb->state |= state;
			break;
		}
	}

}

/* reset all state. */
void
wlc_scb_resetstate(struct scb *scb)
{
	WL_ASSOC(("reset state\n"));
	ASSERT(scb);
	memset(&scb->auth_bsscfg, 0, SCB_BSSCFG_BITSIZE);
	scb->state = 0;
	WL_ASSOC(("wlc_scb : state = %x\n", scb->state));
}

/* initialize per-scb state utilized by rate selection */
void
wlc_scb_rate_init(wlc_info_t *wlc, struct scb *scb)
{
	wlc_rate_init(scb->ratestate, scb->rateset.rates, scb->rateset.count,
		scb->select_rateset.rates, &scb->select_rateset.count,
		0, &scb->txc.txhlen);
}


void
wlc_scb_reinit(wlc_info_t *wlc)
{
	uint prev_count;
	const rateset_t *rs;
	wlcband_t *band;
	struct scb *scb;
	struct scb_iter scbiter;
	/* sanitize any existing scb rates against the current hardware rates */
	FOREACHSCB(&scbiter, wlc->scbstate->scb, scb) {
		prev_count = scb->rateset.count;
		/* Keep only CCK if gmode == GMODE_LEGACY_B */
		band = ((struct scb_info*)scb)->band;
		if (ISGPHY(band->pi) && (band->gmode == GMODE_LEGACY_B)) {
			rs = cck_rates;
		} else {
			rs = &band->hw_rateset;
		}
		wlc_sort_rates(&scb->rateset, rs);
		/* if the count of rates is different, then the rate state
		 * needs to be reinitialized
		 */
		if (scb->rateset.count != prev_count)
			wlc_scb_rate_init(wlc, scb);
	}
}


static INLINE struct scb*
_wlc_scbfind(wlc_info_t *wlc, const struct ether_addr *ea, int bandunit)
{
	int index;
	struct scb_info *scbinfo;

	/* map any multicast address into the single broadcast scb */
	if (ETHER_ISMULTI(ea))
		ea = &ether_bcast;

	index = SCBHASHINDEX(ea->octet);

	/* search for the scb which corresponds to the remote station ea */
	scbinfo = (struct scb_info *)(SCBSTATE_PRIV(wlc->scbstate)->scbhash[bandunit][index]);
	for (; scbinfo; scbinfo = scbinfo->hashnext) {
		if (bcmp((char*)ea, (char*)&(scbinfo->scbpub.ea), ETHER_ADDR_LEN) == 0)
			break;
	}

	return (struct scb *)scbinfo;
}

/* Find station control block corresponding to the remote id */
struct scb *
wlc_scbfind(wlc_info_t *wlc, const struct ether_addr *ea)
{
	return _wlc_scbfind(wlc, ea, wlc->band->bandunit);
}

/*
 * Lookup station control block corresponding to the remote id.
 * If not found, create a new entry.
 */
static INLINE struct scb *
_wlc_scblookup(wlc_info_t *wlc, const struct ether_addr *ea, int bandunit)
{
	int index;
	struct scb *scb;
	scb_module_priv_t *scbstate;

	if ((scb = _wlc_scbfind(wlc, ea, bandunit))) {
		return (scb);
	}

	/* no scb match, allocate one */
	if (!(scb = wlc_scballoc(wlc))) {
		WL_ERROR(("wl%d: wlc_scblookup: out of memory, malloced %d bytes\n",
		          wlc->pub.unit, MALLOCED(wlc->pub.osh)));
		return (NULL);
	}

	/* map any multicast address into the single broadcast scb */
	if (ETHER_ISMULTI(ea))
		ea = &ether_bcast;
	scb->ea = *((struct ether_addr*)ea);

	/* install it in the cache */
	scbstate = SCBSTATE_PRIV(wlc->scbstate);
	scb->bandunit = bandunit;
	index = SCBHASHINDEX(ea->octet);
	((struct scb_info *)scb)->hashnext =
		(struct scb_info *)scbstate->scbhash[bandunit][index];
	scbstate->scbhash[bandunit][index] = scb;

	/* send ofdm rate probe */
	if (!wlc->pub.BSS && wlc->band->gmode && wlc->pub.up)
		wlc_rateprobe(wlc, (struct ether_addr*)ea, scb, WLC_RATEPROBE_RATE);

	return (scb);
}

struct scb *
wlc_scblookup(wlc_info_t *wlc, const struct ether_addr *ea)
{
	return (_wlc_scblookup(wlc, ea, wlc->band->bandunit));
}

struct scb *
wlc_scblookupband(wlc_info_t *wlc, const struct ether_addr *ea, int bandunit)
{
	/* assert that the band is the current band, or we are dual band and it is the other band */
	ASSERT((bandunit == (int)wlc->band->bandunit) ||
	       (NBANDS(wlc) > 1 && bandunit == (int)OTHERBANDUNIT(wlc)));

	return (_wlc_scblookup(wlc, ea, bandunit));
}

/* Get scb from band */
struct scb *
wlc_scbfindband(wlc_info_t *wlc, const struct ether_addr *ea, int bandunit)
{
	/* assert that the band is the current band, or we are dual band and it is the other band */
	ASSERT((bandunit == (int)wlc->band->bandunit) ||
	       (NBANDS(wlc) > 1 && bandunit == (int)OTHERBANDUNIT(wlc)));

	return (_wlc_scbfind(wlc, ea, bandunit));
}

static void
wlc_scb_deauth_sendcomplete(wlc_info_t *wlc, uint txstatus, void *ea_arg)
{
	struct scb *scb;

	/* Is this scb still around */
	if ((scb = wlc_scbfind(wlc, (struct ether_addr*)ea_arg)) == NULL)
	    return;

	/* As long as it made it, complete the state transition */
	wlc_scb_resetstate(scb);

	/* delete pairwise key on deauthentication */
	if (scb->key) {
		WL_WSEC(("wl%d: %s: deauthenticating sta, deleting pairwise"
		         " key for %s\n",
		         wlc->pub.unit, __FUNCTION__,
		         bcm_ether_ntoa(&scb->ea, eabuf)));
		wlc_scb_key_delete(wlc, scb);
	}
}

/* (de)authorize/(de)authenticate single station
 * 'enable' TRUE means authorize, FLASE means deauthorize/deauthenticate
 * 'flag' is AUTHORIZED or AUTHENICATED for the type of operation
 * 'rc' is the reason code for a deauthenticate packet
 */
void
wlc_scb_set_auth(wlc_info_t *wlc, struct scb *scb, bool enable, uint32 flag, int rc)
{
	void *pkt = NULL;

	if (enable) {
		if (flag == AUTHORIZED) {
			wlc_scb_setstatebit(scb, AUTHORIZED);
			scb->flags &= ~SCB_DEAUTH;
		} else {
			wlc_scb_setstatebit(scb, AUTHENTICATED);
		}
	} else {
		if (flag == AUTHORIZED) {

			wlc_scb_clearstatebit(scb, AUTHORIZED);
		} else {

			if (wlc->pub.up && (SCB_AUTHENTICATED(scb))) {
				struct ether_addr *bssid;
				bssid = (scb->flags & SCB_MYAP) ? &wlc->BSSID :
					&wlc->announce_BSSID;
				pkt = wlc_senddeauth(wlc, &scb->ea, bssid, scb, (uint16)rc);
			}
			if (pkt == NULL ||
			    wlc_pkt_callback_register(wlc, wlc_scb_deauth_sendcomplete,
			                              (void *)&scb->ea, pkt))
				WL_ERROR(("wl%d: wlc_scb_set_auth: could not register callback\n",
				          wlc->pub.unit));
		}
	}
	WL_ASSOC(("wl%d: %s %s%s\n", wlc->pub.unit,
		bcm_ether_ntoa(&scb->ea, eabuf),
		(enable ? "" : "de"),
		((flag == AUTHORIZED) ? "authorized" : "authenticated")));
}




void
wlc_scb_sortrates(struct scb *scb)
{
	struct scb_info *scbinfo = (struct scb_info *)scb;
	wlc_sort_rates(&scb->rateset, &scbinfo->band->hw_rateset);
}

void
BCMINITFN(wlc_scblist_validaterates)(wlc_info_t *wlc)
{
	struct scb *scb;
	struct scb_iter scbiter;
	FOREACHSCB(&scbiter, wlc->scbstate->scb, scb) {
		wlc_scb_sortrates(scb);
		if (scb->rateset.count == 0) {
			wlc_scbfree(wlc, scb);
		}
	}
}


/* Give then tx_fn, return the feature id from txmod_fns array.
 * If tx_fn is NULL, 0 will be returned
 * If entry is not found, it's an ERROR!
 */
static INLINE scb_txmod_t
wlc_txmod_fid(wlc_info_t *wlc, txmod_tx_fn_t tx_fn)
{
	scb_txmod_t txmod;

	for (txmod = SCB_START; txmod < SCB_LAST; txmod++)
		if (tx_fn == wlc->txmod_fns[txmod].tx_fn)
			return txmod;
	/* Should not reach here */
	ASSERT(txmod < SCB_LAST);
	return txmod;
}

/* Add a feature to the path. It should not be already on the path and should be configured
 * Does not take care of evicting anybody
 */
static void
wlc_txmod_activate(wlc_info_t *wlc, struct scb *scb, scb_txmod_t fid)
{
	/* Numeric value designating this feature's position in tx_path */
	static const uint txmod_position[SCB_LAST] = {
		0, /* SCB_START */
		1, /* SCB_CRAM */
		1, /* SCB_APPS */
		1, /* SCB_AMSDU */
		2, /* SCB_BA */
		2, /* SCB_AMPDU */
		3, /* SCB_TRANSMIT */
	};

	uint curr_mod_position;
	scb_txmod_t prev, next;
	txmod_info_t curr_mod_info = wlc->txmod_fns[fid];

	ASSERT(SCB_TXMOD_CONFIGURED(scb, fid) &&
	       !SCB_TXMOD_ACTIVE(scb, fid));

	curr_mod_position = txmod_position[fid];

	prev = SCB_START;

	while ((next = wlc_txmod_fid(wlc, scb->tx_path[prev].next_tx_fn)) != 0 &&
	       txmod_position[next] < curr_mod_position)
		prev = next;

	/* next == 0 indicate this is the first addition to the path
	 * it HAS to be SCB_TRANSMIT as it's the one that puts the packet in
	 * txq. If this changes, then assert will need to be removed.
	 */
	ASSERT(next != 0 || fid == SCB_TRANSMIT);
	ASSERT(txmod_position[next] != curr_mod_position);

	SCB_TXMOD_SET(scb, prev, fid);
	SCB_TXMOD_SET(scb, fid, next);

	/* invoke any activate notify functions now that it's in the path */
	if (curr_mod_info.activate_notify_fn)
		curr_mod_info.activate_notify_fn(curr_mod_info.ctx, scb);
}

/* Remove a fid from the path. It should be already on the path
 * Does not take care of replacing it with any other feature.
 */
static void
wlc_txmod_deactivate(wlc_info_t *wlc, struct scb *scb, scb_txmod_t fid)
{
	scb_txmod_t prev, next;
	txmod_info_t curr_mod_info = wlc->txmod_fns[fid];

	/* If not active, do nothing */
	if (!SCB_TXMOD_ACTIVE(scb, fid))
		return;

	/* if deactivate notify function is present, call it */
	if (curr_mod_info.deactivate_notify_fn)
		curr_mod_info.deactivate_notify_fn(curr_mod_info.ctx, scb);

	prev = SCB_START;

	while ((next = wlc_txmod_fid(wlc, scb->tx_path[prev].next_tx_fn))
	       != fid)
		prev = next;

	SCB_TXMOD_SET(scb, prev, wlc_txmod_fid(wlc, scb->tx_path[fid].next_tx_fn));
	scb->tx_path[fid].next_tx_fn = NULL;
}

/* Add the fid to handle packets for this SCB, if allowed */
void
wlc_txmod_config(wlc_info_t *wlc, struct scb *scb, scb_txmod_t fid)
{
	ASSERT(fid < SCB_LAST);

	/* Don't do anything if not yet registered or
	 * already configured
	 */
	if ((wlc->txmod_fns[fid].tx_fn == NULL) ||
	    (SCB_TXMOD_CONFIGURED(scb, fid)))
		return;

	/* Indicate that the feature is configured */
	scb->tx_path[fid].configured = TRUE;

	ASSERT(!SCB_TXMOD_ACTIVE(scb, fid));
	/* Try to activate this feature by adding it to the path
	 * If conflicting features exist in the path, then
	 *     - If this feature is of higher precedence, deactivate other features
	 *     - Else, mark this feature deactivated
	 * Note: When evicting other features, they remain 'configured'
	 */
	switch (fid) {
	case SCB_APPS:
		/* Deactivate CRAM/AMSDU/BA/AMPDU */
		wlc_txmod_deactivate(wlc, scb, SCB_CRAM);
		wlc_txmod_deactivate(wlc, scb, SCB_AMSDU);
		wlc_txmod_deactivate(wlc, scb, SCB_BA);
		wlc_txmod_deactivate(wlc, scb, SCB_AMPDU);
		break;
	case SCB_CRAM:
		/* Don't activate if AMSDU or APPS are active */
		if (SCB_TXMOD_ACTIVE(scb, SCB_AMSDU) ||
		    SCB_TXMOD_ACTIVE(scb, SCB_APPS))
			return;
		break;
	case SCB_AMSDU:
		/* Don't activate if APPS is active */
		if (SCB_TXMOD_ACTIVE(scb, SCB_APPS))
			return;
		/* Deactivate CRAM if on path */
		wlc_txmod_deactivate(wlc, scb, SCB_CRAM);
		break;
	case SCB_BA:
		/* Don't activate if AMPDU is active */
		if (SCB_TXMOD_ACTIVE(scb, SCB_AMPDU))
			return;
		/* Don't activate if APPS is active */
		if (SCB_TXMOD_ACTIVE(scb, SCB_APPS))
			return;
		break;
	case SCB_AMPDU:
		/* Don't activate if APPS is active */
		if (SCB_TXMOD_ACTIVE(scb, SCB_APPS))
			return;
		/* Deactivate BA if on path */
		wlc_txmod_deactivate(wlc, scb, SCB_BA);
		break;
	case SCB_TRANSMIT:
		break;
	case SCB_START:
	case SCB_LAST:
		ASSERT(0);
	};

	wlc_txmod_activate(wlc, scb, fid);
}

/* Remove the feature to handle packets for this SCB.
 * If just configured but not in path, just marked unconfigured
 * If in path, feature is removed and, if applicable, replaced by any other feature
 */
void
wlc_txmod_unconfig(wlc_info_t *wlc, struct scb *scb, scb_txmod_t fid)
{
	ASSERT(fid < SCB_LAST && fid != SCB_TRANSMIT);

	if (!SCB_TXMOD_CONFIGURED(scb, fid))
		return;

	scb->tx_path[fid].configured = FALSE;

	/* Nothing to do if not active */
	if (!SCB_TXMOD_ACTIVE(scb, fid))
		return;

	wlc_txmod_deactivate(wlc, scb, fid);

	/* Restore any other features to the path */
	/* APPS unconfigured  => CRAM/AMSDU/BA can be active
	 * AMSDU unconfigured => CRAM can be active
	 */
	switch (fid) {
	case SCB_APPS:
		/* Activate AMSDU/BA/CRAM if configured
		 * Order below helps reduce redundant operations
		 * if both CRAM/AMSDU are configured
		 */
		if (SCB_TXMOD_CONFIGURED(scb, SCB_AMSDU))
			wlc_txmod_activate(wlc, scb, SCB_AMSDU);
		else if (SCB_TXMOD_CONFIGURED(scb, SCB_CRAM))
			wlc_txmod_activate(wlc, scb, SCB_CRAM);
		if (SCB_TXMOD_CONFIGURED(scb, SCB_BA))
			wlc_txmod_activate(wlc, scb, SCB_BA);
		if (SCB_TXMOD_CONFIGURED(scb, SCB_AMPDU))
			wlc_txmod_activate(wlc, scb, SCB_AMPDU);
		break;
	case SCB_AMSDU:
		/* Activate CRAM if configured and APPS is not active */
		if (SCB_TXMOD_CONFIGURED(scb, SCB_CRAM) &&
		    !SCB_TXMOD_ACTIVE(scb, SCB_APPS))
			wlc_txmod_activate(wlc, scb, SCB_CRAM);
		break;
	case SCB_CRAM:
	case SCB_BA:
	case SCB_AMPDU:
	case SCB_TRANSMIT:
	case SCB_START:
	case SCB_LAST:
		break;
	}
}
